__author__ = 'Sergei F. Kliver'

from KRATER.Routines.Math import MathRoutines
from KRATER.Routines.File import FileRoutines
from KRATER.Routines.Sequence import SequenceRoutines
from KRATER.Routines.Matplotlib import MatplotlibRoutines


MathRoutines = MathRoutines()
FileRoutines = FileRoutines()
SequenceRoutines = SequenceRoutines()
MatplotlibRoutines = MatplotlibRoutines()
